#ifndef EXPRESSIONS_H
#define EXPRESSIONS_H

#include "config.h"




#include <array>
#include <algorithm>
#include <initializer_list>
#include <type_traits>
#include <iostream>

template <std::size_t N> class math_vector;
template <typename T> struct plus_op;
template <typename T> struct minus_op;
template <typename T> struct mul_op;
template <typename T> struct cross_op;

template <
  typename LeftExpr,
  typename BinaryOp,
  typename RightExpr
>
class math_vector_expr
{
  public:
    typedef typename std::remove_reference<LeftExpr>::type le_type;
    typedef typename le_type::value_type  le_value_type;
    static const std::size_t le_size = le_type::size;

    math_vector_expr() = delete;

    math_vector_expr(LeftExpr l, RightExpr r) :
      l_(std::forward<LeftExpr>(l)),
      r_(std::forward<RightExpr>(r))
    {}

    // Prohibit copying...
    math_vector_expr(math_vector_expr const&) = delete;
    math_vector_expr& operator =(math_vector_expr const&) = delete;

    // Allow moves...
    math_vector_expr(math_vector_expr&&) = default;
    math_vector_expr& operator =(math_vector_expr&&) = default;

    template <typename RE>
    auto operator +(RE&& re) const
    {
      return
        math_vector_expr<
          math_vector_expr<LeftExpr,BinaryOp,RightExpr> const&,
          plus_op<le_value_type>,
          decltype(std::forward<RE>(re))
        >(*this, std::forward<RE>(re))
      ;
    }

    template <typename RE>
    auto operator -(RE&& re) const {
      return
        math_vector_expr<
          math_vector_expr<LeftExpr,BinaryOp,RightExpr> const&,
          minus_op<le_value_type>,
          decltype(std::forward<RE>(re))
        >(*this, std::forward<RE>(re));
    }

    auto operator *(SIMD re) const
    {
      return
        math_vector_expr<
          math_vector_expr<LeftExpr,BinaryOp,RightExpr> const&,
          mul_op<le_value_type>,
          le_type
        >(
          *this,
          le_type{{re,re,re}}   ///@todo: arbitrary number of dimensions
        )
      ;
    }


    auto le() -> typename std::add_lvalue_reference<LeftExpr>::type
      { return l_; }

    auto le() const -> typename std::add_lvalue_reference<
                            typename std::add_const<LeftExpr>::type
                       >::type
      { return l_; }

    auto re() -> typename std::add_lvalue_reference<RightExpr>::type
      { return r_; }

    auto re() const -> typename std::add_lvalue_reference<
                            typename std::add_const<RightExpr>::type
                       >::type
      { return r_; }

    auto operator [](std::size_t index) const {
      return BinaryOp::apply(le()[index], re()[index]);
    }

  private:
    LeftExpr l_;
    RightExpr r_;
};

//template <
//  typename LeftExpr,
//  typename RightExpr
//>
//class math_vector_expr<LeftExpr, cross_op<Real>, RightExpr>
//{
//    auto operator [](std::size_t index) const
//    {
//      switch(index){
//      case 0:
//          return le()[1]*re()[2]-le()[2]*re()[1];
//      case 1:
//          return le()[2]*re()[0]-le()[0]*re()[2];
//      case 2:
//          return le()[0]*re()[1]-le()[1]*re()[0];
//      }
//    }
//};

template <typename T>
struct plus_op {
  static T apply(T const& a, T const& b)  { return a + b; }
  static T apply(T&& a, T const& b)       { a += b; return std::move(a); }
  static T apply(T const& a, T&& b)       { b += a; return std::move(b); }
  static T apply(T&& a, T&& b)            { a += b; return std::move(a); }
};

template <typename T>
struct minus_op {
  static T apply(T const& a, T const& b)  { return a - b; }
  static T apply(T&& a, T const& b)       { a -= b; return std::move(a); }
  static T apply(T const& a, T&& b)       { b -= a; return std::move(b); }
  static T apply(T&& a, T&& b)            { a -= b; return std::move(a); }
};

template <typename T>
struct mul_op {
  static T apply(T const& a, SIMD const& b)  { return a * b; }
  static T apply(T&& a, SIMD const& b)       { a *= b; return std::move(a); }
  static T apply(T&& a, SIMD&& b)            { a *= b; return std::move(a); }
};

template <typename T>
struct div_op {
  static T apply(T const& a, SIMD const& b)  { return a / b;  }
  static T apply(T&& a, SIMD  const& b)      { a /= b; return std::move(a); }
  static T apply(T&& a, SIMD&& b)            { a /= b; return std::move(a); }
};

template <typename T>
struct cross_op {
    typedef typename T::type value_type;
};

template <std::size_t N>
class math_vector
{
  using impl_type = std::array<SIMD, N>;

  public:
    typedef typename impl_type::value_type value_type;
    static const std::size_t size = N;

    math_vector()
    {
      for (std::size_t i=0; i<N; ++i)
          v_[i] = _mm_set1_ps(0);
    }

    math_vector(math_vector const& mv) noexcept
    {
      using namespace std;
      copy(begin(mv.v_), end(mv.v_), begin(v_));
    }

    math_vector(math_vector&& mv) noexcept
    {
      using namespace std;
      move(begin(mv.v_), end(mv.v_), begin(v_));
    }

    math_vector(std::initializer_list<typename impl_type::value_type> l)
    {
      using namespace std;
      copy(begin(l), end(l), begin(v_));
    }

    math_vector& operator =(math_vector const& mv) noexcept
    {
      using namespace std;
      copy(begin(mv.v_), end(mv.v_), begin(v_));
      return *this;
    }

    math_vector& operator =(math_vector&& mv) noexcept
    {
      using namespace std;
      move(begin(mv.v_), end(mv.v_), begin(v_));
      return *this;
    }

    ~math_vector(){}

    void swap(math_vector& mv)
    {
      using namespace std;
      for (std::size_t i = 0; i<N; ++i)
        swap(v_[i], mv[i]);
    }

    auto operator [](std::size_t index) const
      -> typename impl_type::value_type const&
    {
      return v_[index];
    }

    auto operator [](std::size_t index)
      -> typename impl_type::value_type&
    {
      return v_[index];
    }

    math_vector& operator +=(math_vector const& b)
    {
      for (std::size_t i = 0; i<N; ++i)
        v_[i] += b[i];
      return *this;
    }

    template <typename LE, typename Op, typename RE>
    math_vector(math_vector_expr<LE,Op,RE>&& mve)
    {
      for (std::size_t i = 0; i < N; ++i)
        v_[i] = mve[i];
    }

    template <typename RightExpr>
    math_vector& operator =(RightExpr&& re)
    {
      for (std::size_t i = 0; i<N; ++i)
        v_[i] = re[i];
      return *this;
    }

    template <typename RightExpr>
    math_vector& operator +=(RightExpr&& re)
    {
      for (std::size_t i = 0; i<N; ++i)
        v_[i] += re[i];
      return *this;
    }

    template <typename RightExpr>
    auto operator +(RightExpr&& re) const
    {
      return
        math_vector_expr<
          math_vector const&,
          plus_op<typename impl_type::value_type>,
          decltype(std::forward<RightExpr>(re))
        >(
          *this,
          std::forward<RightExpr>(re)
        )
      ;
    }

    template <typename RightExpr>
    math_vector& operator -=(RightExpr&& re)
    {
      for (std::size_t i = 0; i<N; ++i)
        v_[i] -= re[i];
      return *this;
    }

    template <typename RightExpr>
    auto operator -(RightExpr&& re) const
    {
      return
        math_vector_expr<
          math_vector const&,
          minus_op<typename impl_type::value_type>,
          decltype(std::forward<RightExpr>(re))
        >(
          *this,
          std::forward<RightExpr>(re)
        )
      ;
    }


    auto operator *(SIMD re) const
    {
      return
        math_vector_expr<
          math_vector const&,
          mul_op<typename impl_type::value_type>,
          impl_type
        >(
          *this,
          impl_type{{re,re,re}}   ///@todo: arbitrary number of dimensions, also - will it create a temporary?
        )
      ;
    }



  private:
    impl_type v_;
};

// Throws a warning for pre-C++14 dialects, because it cannot auto-deduce the return type
template <std::size_t N>
inline auto operator*(SIMD a, math_vector<N>& b)
{ return b * a; }

//template <typename LE, typename Op, typename RE>
//inline auto operator*(SIMD a, math_vector_expr<LE, Op, RE> b)
//{ return b * a; }

template <std::size_t N>
inline void swap(math_vector<N>& a, math_vector<N>& b)
{ a.swap(b); }


template <std::size_t N>
std::ostream& operator <<(std::ostream& os, math_vector<N> const& mv)
{
  os << '(';
  for (std::size_t i = 0; i < N; ++i)
    os << mv[i] << ((i+1 != N) ? ',' : ')');
  return os;
}


#endif
